#include "precompile.h"
#include "solver.h"

using namespace std;

const int directions[2][4] = {{0, 0, 1, -1}, {1, -1, 0, 0}};
Position my_queue[MAX*MAX];

int Bfs(int row, int col, int visited[MAX][MAX], int etiket, const Board& board) {

	int start = 0, end = 0;
	int result = 0;

	my_queue[end++] = Position(row, col);

	while (start < end) {
		Position curr_pos(my_queue[start++]);

		if (visited[curr_pos.first][curr_pos.second] == etiket) {
			continue;
		}

		visited[curr_pos.first][curr_pos.second] = etiket;
		++result;

		for (int dir = 0; dir < 4; ++dir) {
			int x = curr_pos.first + directions[0][dir];
			int y = curr_pos.second + directions[1][dir];

			if (x < 0 || x >= (int)board.size() || y < 0 || y >= (int)board[x].size()) {
				continue;
			}

			if (board[row][col] != board[x][y]) {
				continue;
			}
			my_queue[end++] = Position(x, y);
		}
	}

	return result;
}

int Bfs(int row, int col, vector<vector<bool> >& visited, const Board& board) {
	queue<Position> q;
	int result = 0;

	q.push(Position(row, col));

	while (!q.empty()) {
		Position curr_pos(q.front());
		q.pop();

		if (visited[curr_pos.first][curr_pos.second]) {
			continue;
		}

		visited[curr_pos.first][curr_pos.second] = true;
		++result;

		for (int dir = 0; dir < 4; ++dir) {
			int x = curr_pos.first + directions[0][dir];
			int y = curr_pos.second + directions[1][dir];

			if (x < 0 || x >= (int)board.size() || y < 0 || y >= (int)board[x].size()) {
				continue;
			}

			if (board[row][col] != board[x][y]) {
				continue;
			}
			q.push(Position(x, y));
		}
	}

	return result;
}

int Bfs(int row, int col, int visited[MAX][MAX], int etiket, const Board& board, const Individ& individ) {

	int start = 0, end = 0;
	int result = 0;

	my_queue[end++] = Position(row, col);

	int curr_color = -1;
	int real_pos_row_ind;
	int real_pos_col_ind;

	if (individ.type == EVOLVE_ROW) {
		real_pos_row_ind = (row + individ.colShifts[col]) % board.size();
		real_pos_col_ind = (col + individ.rowShifts[real_pos_row_ind]) % board[0].size();
	}
	else {
		real_pos_col_ind = (col + individ.rowShifts[row]) % board[0].size();
		real_pos_row_ind = (row + individ.colShifts[real_pos_col_ind]) % board.size();
	}

	curr_color = board[real_pos_row_ind][real_pos_col_ind];


	while (start < end) {
		Position curr_pos(my_queue[start++]);

		if (visited[curr_pos.first][curr_pos.second] == etiket) {
			continue;
		}

		visited[curr_pos.first][curr_pos.second] = etiket;
		++result;

		for (int dir = 0; dir < 4; ++dir) {
			int x = curr_pos.first + directions[0][dir];
			int y = curr_pos.second + directions[1][dir];

			if (x < 0 || x >= (int)board.size() || y < 0 || y >= (int)board[x].size()) {
				continue;
			}

			int real_row_ind;
			int real_col_ind;

			if (individ.type == EVOLVE_ROW) {
				real_row_ind = (x + individ.colShifts[y]) % board.size();
				real_col_ind = (y + individ.rowShifts[real_row_ind]) % board[0].size();
			}
			else {
				real_col_ind = (y + individ.rowShifts[x]) % board[0].size();
				real_row_ind = (x + individ.colShifts[real_col_ind]) % board.size();
			}

			if (curr_color != board[real_row_ind][real_col_ind]) {
				continue;
			}
			my_queue[end++] = Position(x, y);
		}
	}

	return result;
}


int GetScore(const Input& input, const Solution& s) {
	int result = 0;
	Board board(input.board);
	
	vector<int> temp;
	for (size_t oper_ind = 0; oper_ind < s.operations.size(); ++oper_ind) {
		const Operation& operation(s.operations[oper_ind]);
		if (operation.type == ROW_SHIFT) {
			int row_ind = operation.index;
			temp = board[row_ind];
			for (int col_index = 0; col_index < input.M; ++col_index) {
				size_t curr_col = (col_index + operation.times) % input.M;
				board[row_ind][col_index] = temp[curr_col];
			}
		}
		else {
			int col_ind = operation.index;
			temp.resize(input.N);
			for (int r_ind = 0; r_ind < input.N; ++r_ind) {
				temp[r_ind] = board[r_ind][col_ind];
			}

			for (int row_index = 0; row_index < input.N; ++row_index) {
				size_t curr_row = (row_index + operation.times) % input.N;
				board[row_index][col_ind] = temp[curr_row];
			}
		}
	}

	vector<vector<bool> > visited(input.N, vector<bool>(input.M, false));
	for (int row_ind = 0; row_ind < input.N; ++row_ind) {
		for (int col_ind = 0; col_ind < input.M; ++col_ind) {
			if (visited[row_ind][col_ind]) {
				continue;
			}
			int bfs_res = Bfs(row_ind, col_ind, visited, board);
			result += bfs_res * bfs_res;
		}
	}
	return result;
}

Individ::Individ() :
		score(-1) {
	/*memset(this->rowShifts, 0, sizeof(this->rowShifts));
	memset(this->colShifts, 0, sizeof(this->colShifts));*/
}

bool Individ::operator < (const Individ& rhs) const {
	assert(this->score != -1 && rhs.score != -1);
	return this->score > rhs.score;
}

Solver::Solver() :
	populationSize(0),
	input(NULL),
	mutationTime(0),
	evaluationTime(0),
	etiket(0),
	mutations(0) {
		memset(visited, 0, sizeof(visited));
}

void Solver::Clear() {
	this->populationSize = 0;
	this->input = NULL;
	this->mutationTime = 0;
	this->evaluationTime = 0;
	this->mutations = 0;
}


bool Solver::Solve(const Input& input, Solution* result) {
	this->input = &input;
	int initial_population = MAX_INDIVIDS / (this->input->N * this->input->M);

	// Init test data.
	this->testBoard.resize(this->input->N);
	for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
		this->testBoard[row_ind].resize(this->input->M);
	}

	GenerateInitialPopulation(initial_population);
	clock_t beg = clock();
	const int AGE_TIME_CAP = TIME_TO_EVOLVE; 

	int CURR_LEVEL_CUT = 1 + int(sqrt(double(this->populationSize)) + 1e-3);
	int CURR_LEVEL_RANDOM_FACTOR = CURR_LEVEL_CUT / 10;

	while (true) {
		sort(&this->population[0], &this->population[this->populationSize]);

		clock_t curr_time = clock() - beg;
		if (curr_time > AGE_TIME_CAP) {
			break;
		}
		this->populationSize = min(CURR_LEVEL_CUT, (int)this->populationSize);
		int pop_cnt = this->populationSize;

		for (int individ = 0; individ < pop_cnt; ++individ) {
			for (int individ2 = individ + 1; individ2 < pop_cnt; ++individ2) {
				Mutate(this->population[individ], this->population[individ2], 
					&this->population[this->populationSize]);

				++this->populationSize;
				if (this->populationSize == MAX_POSSIBLE_INDIVIDS) {
					break;
				}
			}
		}

		/*for (size_t r_ind = 0; r_ind < CURR_LEVEL_RANDOM_FACTOR && this->populationSize < MAX_POSSIBLE_INDIVIDS; ++r_ind) {
			this->population[this->populationSize] = this->population[r_ind];
			Randomize(&this->population[this->populationSize]);
			++this->populationSize;
		}*/
	}

	int score = this->population[0].score;
	this->population[0].type = (this->population[0].type == EVOLVE_COL ? EVOLVE_ROW : EVOLVE_COL);
	Evaluate(&this->population[0]);
	if (score > this->population[0].score) {
		this->population[0].type = (this->population[0].type == EVOLVE_COL ? EVOLVE_ROW : EVOLVE_COL);
	}
	ToSolution(this->population[0], result);
#ifdef TEST
	printf("Cutoff score: %d\n", CURR_LEVEL_CUT);
	PrintTimes();
#endif

	return true;
}

void Solver::Randomize(Individ* individ) {
	for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
		if (rand() % 3 != 1) {
			individ->rowShifts[row_ind] = 1 + (unsigned)rand() % 
				(this->input->M - 1);
		}
	}

	for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
		if (rand() % 3 != 1) {
			individ->colShifts[col_ind] = 1 + (unsigned)rand() %
				(this->input->N - 1);
		}
	}

	Evaluate(individ);
}

void Solver::Mutate(const Individ& lhs, const Individ& rhs, Individ* result) {
#ifdef TEST
	clock_t beg = clock();
	++this->mutations;
#endif

	for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
		result->rowShifts[row_ind] = (rand() % 2 == 1) ? lhs.rowShifts[row_ind] :
			rhs.rowShifts[row_ind];
	}
	
	for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
		result->colShifts[col_ind] = (rand() % 2 == 1) ? lhs.colShifts[col_ind] :
			rhs.colShifts[col_ind];
	}

	Evaluate(result);

#ifdef TEST
	//Solution s;
	//ToSolution(*result, &s);
	//assert(GetScore(*this->input, s) == result->score);
#endif

#ifdef TEST
	this->mutationTime += clock() - beg;
#endif
}

void Solver::Evaluate(Individ* individ) {
#ifdef TEST
	clock_t beg = clock();
#endif
	individ->type = EVOLVE_ROW;
	individ->score = 0;
	++this->etiket;
	for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
		for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
			if (this->visited[row_ind][col_ind] == this->etiket) {
				continue;
			}
			
			int curr_score = Bfs(row_ind, col_ind, this->visited, this->etiket, this->input->board, *individ);
			individ->score += curr_score * curr_score;
		}
	}

	if (this->input->N * this->input->M <= 128) {
		int col_score = 0;
		individ->type = EVOLVE_COL;
		//ToTestBoard(*individ);
		++this->etiket;
		for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
			for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
				if (this->visited[row_ind][col_ind] == this->etiket) {
					continue;
				}

				int curr_score = Bfs(row_ind, col_ind, this->visited, this->etiket, this->input->board, *individ);
				col_score += curr_score * curr_score;
			}
		}

		if (individ->score < col_score) {
			individ->score = col_score;
		}
		else {
			individ->type = EVOLVE_ROW;
		}
	}

#ifdef TEST
	this->evaluationTime += clock() - beg;
#endif
}

void Solver::ToSolution(const Individ& individ, Solution* result) {
	if (individ.type == EVOLVE_ROW) {
		for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
			if (individ.rowShifts[row_ind] != 0) {
				result->operations.push_back(
					Operation(ROW_SHIFT, row_ind, individ.rowShifts[row_ind]));
			}
		}
	}
	
	for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
		if (individ.colShifts[col_ind] != 0) {
			result->operations.push_back(
				Operation(COL_SHIFT, col_ind, individ.colShifts[col_ind]));
		}
	}

	if (individ.type == EVOLVE_COL) {
		for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
			if (individ.rowShifts[row_ind] != 0) {
				result->operations.push_back(
					Operation(ROW_SHIFT, row_ind, individ.rowShifts[row_ind]));
			}
		}
	}

#ifdef TEST
	int score = GetScore(*this->input, *result);
	assert(individ.score == score);
#endif
}

void Solver::GenerateInitialPopulation(int population_size) {
#ifdef TEST
	clock_t begin = clock();
#endif
	population_size = min(population_size, MAX_POSSIBLE_INDIVIDS);
	for (int individ_ind = 0; individ_ind < population_size; ++individ_ind) {
		for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
			if (rand() % 3 != 1) {
				this->population[individ_ind].rowShifts[row_ind] = 1 + (unsigned)rand() % 
					(this->input->M - 1);
			}
			else {
				this->population[individ_ind].rowShifts[row_ind] = 0;
			}
		}

		for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
			if (rand() % 3 != 1) {
				this->population[individ_ind].colShifts[col_ind] = 1 + (unsigned)rand() %
					(this->input->N - 1);
			}
			else {
				this->population[individ_ind].colShifts[col_ind] = 0;
			}
		}
		Evaluate(&this->population[individ_ind]);
#ifdef TEST
		//Solution s;
		//ToSolution(this->population[individ_ind], &s);
		//assert(GetScore(*this->input, s) == this->population[individ_ind].score);
#endif

	}
	this->populationSize = population_size;
#ifdef TEST
	printf("Time to generate randoms: %.3lf sec\n", (clock() - begin) / (double)CLOCKS_PER_SEC);
#endif
}

void Solver::ToTestBoard(const Individ& individ) {
	EvolutionPart type = individ.type;
	for (int row_ind = 0; row_ind < this->input->N; ++row_ind) {
		for (int col_ind = 0; col_ind < this->input->M; ++col_ind) {
			int real_row_ind;
			int real_col_ind;

			if (type == EVOLVE_ROW) {
				real_row_ind = (row_ind + individ.colShifts[col_ind]) % this->input->N;
				real_col_ind = (col_ind + individ.rowShifts[real_row_ind]) % this->input->M;
			}
			else {
				real_col_ind = (col_ind + individ.rowShifts[row_ind]) % this->input->M;
				real_row_ind = (row_ind + individ.colShifts[real_col_ind]) % this->input->N;
			}
			this->testBoard[row_ind][col_ind] = this->input->board[real_row_ind][real_col_ind];
		}
	}
}

void Solver::PrintTimes() const {
	printf("Time spent on mutation: %.3lf sec\n", this->mutationTime / (double)CLOCKS_PER_SEC);
	printf("Mutation attempts: %u\n", this->mutations);
	printf("Time spent on evaluation: %.3lf sec\n", this->evaluationTime / (double)CLOCKS_PER_SEC);
}
