#pragma once

#include "interface_fwd.h"

#define MAX 128
#define MAX_POSSIBLE_INDIVIDS 10000

int GetScore(const Input& input, const Solution& s);

enum EvolutionPart {
	EVOLVE_ROW,
	EVOLVE_COL
};

struct Individ {
	Individ();

	bool operator < (const Individ& rhs) const;

	int score;
	EvolutionPart type;
	char rowShifts[128];
	char colShifts[128];
};

class Solver {
public:
	Solver();
	bool Solve(const Input& input, Solution* result);
	void Clear();
private:
	void GenerateInitialPopulation(int population_size);
	void Mutate(const Individ& lhs, const Individ& rhs, Individ* result);
	void Evaluate(Individ* individ);
	void ToSolution(const Individ& individ, Solution* result);
	void ToTestBoard(const Individ& individ);
	void Randomize(Individ* input);

	void PrintTimes() const;
private:
	Individ population[MAX_POSSIBLE_INDIVIDS];
	size_t populationSize;
	const Input* input;
	int visited[MAX][MAX];
	Board testBoard;
	int etiket;

	clock_t mutationTime;
	clock_t evaluationTime;
	unsigned mutations;
};

